import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Sequence,
    Choice,
    Keyword,
    Token,
    Repeat,
)  # nopep8


class TestRepeat(unittest.TestCase):

    def test_repeat(self):
        k_hi = Keyword('hi')
        repeat = Repeat(k_hi)
        grammar = create_grammar(repeat)

        self.assertEqual(repeat.min, 0)
        self.assertEqual(repeat.max, None)
        self.assertTrue(grammar.parse('hi hi hi').is_valid)
        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertTrue(grammar.parse('').is_valid)
        self.assertFalse(grammar.parse('hihi').is_valid)
        self.assertFalse(grammar.parse('ha,').is_valid)
        self.assertEqual(
            grammar.parse('hi.').as_str(),
            'error at position 2, expecting: end_of_statement or hi'
        )
        self.assertEqual(
            grammar.parse('hi.').as_str(translate=lambda elem: ''),
            'error at position 2'
        )

    def test_repeat_all_options(self):
        k_hi = Keyword('hi')
        repeat = Repeat(k_hi, mi=1, ma=3)
        grammar = create_grammar(repeat)

        self.assertEqual(repeat.min, 1)
        self.assertEqual(repeat.max, 3)
        self.assertTrue(grammar.parse('hi hi hi').is_valid)
        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertFalse(grammar.parse('').is_valid)
        self.assertFalse(grammar.parse('hi hi hi hi').is_valid)
        self.assertEqual(
            grammar.parse('hi hi hi hi hi.').as_str(),
            'error at position 8, expecting: end_of_statement'
        )
        self.assertEqual(
            grammar.parse('hi.').as_str(),
            'error at position 2, expecting: end_of_statement or hi'
        )
        self.assertEqual(
            grammar.parse('').as_str(),
            'error at position 0, expecting: hi'
        )


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Keyword,
    Ref,
)  # nopep8


class TestRef(unittest.TestCase):

    def test_ref(self):
        ref = Ref()
        k_hi = Keyword('HI')
        ref = k_hi
        grammar = create_grammar(ref)

        self.assertTrue(grammar.parse('HI').is_valid)
        self.assertFalse(grammar.parse('hi').is_valid)
        self.assertFalse(grammar.parse('').is_valid)
        self.assertEqual(
            str(grammar.parse('ha')),
            'error at position 0, expecting: HI',
        )


if __name__ == '__main__':
    unittest.main()



"""Element and NamedElement Class.

These are the base classes used for all other elements.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
"""
import typing as t


def camel_case(s: str) -> str:
    return ''.join(
        p[0].upper() + p[1:] if n else p
        for n, p in enumerate(s.split('_')))


def cap_case(s: str) -> str:
    return ''.join(p[0].upper() + p[1:] for p in s.split('_') if p)


_cref = None


def c_export(func):

    def wrapper(self, c_indent, indent, enums, ref=None):
        global _cref
        elem, _cref = self if _cref is None else _cref, ref
        gid = getattr(elem, 'name', getattr(elem, '_name', 'CLERI_NONE'))
        if gid != 'CLERI_NONE':
            gid = 'CLERI_GID_{}'.format(gid.upper())
            enums.add(gid)

        return func(self, c_indent, indent, enums, gid)

    return wrapper


def go_export(func):

    def wrapper(self, go_indent, indent, enums):
        gid = getattr(self, 'name', getattr(self, '_name', 'NoGid'))
        if gid != 'NoGid':
            gid = 'Gid{}'.format(cap_case(gid))
            enums.add(gid)
        return func(self, go_indent, indent, enums, gid)

    return wrapper


def java_export(func):

    def wrapper(self, java_indent, indent, enums, classes):
        classes.add('jleri.{}'.format(self.__class__.__name__.lstrip('_')))
        gid = getattr(self, 'name', getattr(self, '_name', None))
        if gid is not None:
            gid = gid.upper()
            enums.add(gid)
        return func(self, java_indent, indent, enums, classes, gid)

    return wrapper


class Element:

    __slots__ = tuple()

    name: t.Optional[str]

    @staticmethod
    def _validate_element(element: t.Union[str, 'Element']) -> 'Element':
        if isinstance(element, str):
            return Token(element)
        if isinstance(element, Element):
            return element
        raise TypeError(
            'Expecting an element or string '
            'but received type: {}'.format(type(element)))

    @classmethod
    def _validate_elements(cls, elements) -> t.List['Element']:
        return [cls._validate_element(elem) for elem in elements]


class NamedElement(Element):

    __slots__ = ('name',)

    def _export_js(
            self,
            js_indent: int,
            indent: int,
            classes,
            cname) -> str:
        classes.add(self.__class__.__name__.lstrip('_'))
        if hasattr(self, 'name') and indent > 0:
            return '{}.{}'.format(cname, self.name) if cname else self.name
        indent = 0 if indent < 0 else 1 if indent == 0 else indent
        return self._run_export_js(js_indent, indent, classes, cname)

    def _export_js_elements(self, js_indent, indent, classes, cname):
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=js_indent * new_indent,
            elem=elem._export_js(
                js_indent,
                new_indent, classes, cname)) for elem in self._elements])
        return '{class_name}(\n{value}\n{indent})'.format(
            class_name=self.__class__.__name__.lstrip('_'),
            value=value,
            indent=js_indent * indent)

    def _run_export_js(self, js_indent, indent, classes, cname):
        return 'not_implemented'

    def _export_py(self, py_indent, indent, classes):
        classes.add(self.__class__.__name__.lstrip('_'))
        if hasattr(self, 'name') and indent:
            return self.name
        return self._run_export_py(py_indent, indent or 1, classes)

    def _export_py_elements(self, py_indent, indent, classes):
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=py_indent * new_indent,
            elem=elem._export_py(
                py_indent,
                new_indent, classes)) for elem in self._elements])
        return '{class_name}(\n{value}\n{indent})'.format(
            class_name=self.__class__.__name__.lstrip('_'),
            value=value,
            indent=py_indent * indent)

    def _run_export_py(self, py_indent, indent, classes):
        return 'not_implemented'

    @c_export
    def _export_c(self, c_indent, indent, enums, gid):
        if hasattr(self, 'name') and indent:
            return self.name
        return self._run_export_c(c_indent, indent or 1, enums)

    @c_export
    def _export_c_elements(self, c_indent, indent, enums, gid):
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=c_indent * new_indent,
            elem=elem._export_c(
                c_indent,
                new_indent,
                enums)) for elem in self._elements])
        return \
            'cleri_{class_name}(\n{gid},\n{num},\n{value}\n{indent})'.format(
                class_name=self.__class__.__name__.lstrip('_').lower(),
                gid='{indent}{gid}'.format(
                    indent=c_indent * (indent + 1),
                    gid=gid),
                num='{indent}{num}'.format(
                    indent=c_indent * (indent + 1),
                    num=len(self._elements)),
                value=value,
                indent=c_indent * indent)

    def _run_export_c(self, c_indent, indent, enums):
        return 'not_implemented'

    @go_export
    def _export_go(self, go_indent, indent, enums, gid):
        if hasattr(self, 'name') and indent:
            return camel_case(self.name)
        return self._run_export_go(go_indent, indent or 1, enums)

    @go_export
    def _export_go_elements(self, go_indent, indent, enums, gid):
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=go_indent * new_indent,
            elem=elem._export_go(
                go_indent,
                new_indent,
                enums)) for elem in self._elements])
        return '{class_name}(\n{gid},\n{value},\n{indent})'.format(
            class_name='goleri.New' + self.__class__.__name__.lstrip('_'),
            gid='{indent}{gid}'.format(
                indent=go_indent * (indent + 1),
                gid=gid),
            value=value,
            indent=go_indent * indent)

    def _run_export_go(self, go_indent, indent, enums):
        return 'not_implemented'

    @java_export
    def _export_java(self, java_indent, indent, enums, classes, gid):
        if hasattr(self, 'name') and indent > 0:
            return self.name.upper()
        return self._run_export_java(
            java_indent, abs(indent) or 1, enums, classes)

    @java_export
    def _export_java_elements(
            self, java_indent, indent, enums, classes, gid):
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=java_indent * new_indent,
            elem=elem._export_java(
                java_indent,
                new_indent,
                enums,
                classes)) for elem in self._elements])
        return '{class_name}({gid}\n{value}\n{indent})'.format(
            class_name='new ' + self.__class__.__name__.lstrip('_'),
            gid='' if gid is None else '\n{indent}Ids.{gid},'.format(
                indent=java_indent * (indent + 1),
                gid=gid),
            value=value,
            indent=java_indent * indent)

    def _run_export_java(self, java_indent, indent, enums, classes):
        return 'not_implemented'

# Added this import to the bottom to prevent circular import cycle.
# Note: usually this is bad design but in this case we do want class
#       inheritance which allows us to create a new class Token which
#       is sub-classed from the 'NamedElement' class.


from .token import Token  # nopep8



"""Node class.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
"""
import typing as t
if t.TYPE_CHECKING:
    from .elements import Element


class Node:

    __slots__ = ('element', '_string', 'start', 'end', 'children')

    def __init__(
            self,
            element: 'Element',
            string: str,
            start: int,
            end: t.Optional[int] = None):
        self.element = element
        self.start = start
        self.end = end
        self._string = string
        self.children: t.List['Node'] = []

    @property
    def string(self) -> str:
        return self._string[self.start:self.end]



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Keyword,
)  # nopep8
from examples import JsonGrammar  # nopep8


class TestJsonLang(unittest.TestCase):

    def test_json(self):
        grammar = JsonGrammar()
        # should be true
        self.assertTrue(grammar.parse("""
            {
                "hoi \\"Iris\\"": [
                    5,
                    -0.5,
                    {
                        "a": true,
                        "b": false
                    }
                ]
            }""").is_valid)

        # keys must be strings
        self.assertFalse(grammar.parse('{5: true}').is_valid)

        # floats should start with 0-9.
        self.assertFalse(grammar.parse('{"a": .5}').is_valid)

        # wrong escaping
        self.assertFalse(grammar.parse('{"a\\": null}').is_valid)

        # empty strings are not allowed
        self.assertFalse(grammar.parse('').is_valid)


if __name__ == '__main__':
    unittest.main()



"""end_of_statement Variable.

end_of_statement is an instance of _EndOfStatement and will be added to
an 'expecting' in a node result when an 'End of Statement' is possible.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
"""


class _EndOfStatement:

    def __repr__(self):
        return 'end_of_statement'


end_of_statement = _EndOfStatement()



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Tokens,
)  # nopep8


class TestTokens(unittest.TestCase):

    def test_tokens(self):
        spaced = '== != >= <= > <'
        tokens = Tokens('== > !=  <  >=  <= ')
        grammar = create_grammar(tokens)

        self.assertEqual(spaced, str(tokens))
        self.assertTrue(grammar.parse('==').is_valid)
        self.assertTrue(grammar.parse('<=').is_valid)
        self.assertTrue(grammar.parse('>').is_valid)
        self.assertFalse(grammar.parse('').is_valid)
        self.assertFalse(grammar.parse('=').is_valid)
        self.assertEqual(
            str(grammar.parse('')),
            'error at position 0, expecting: == != >= <= > <'
        )


if __name__ == '__main__':
    unittest.main()



"""setup.py

Upload to PyPI, Thx to: http://peterdowns.com/posts/first-time-with-pypi.html

python3 setup.py sdist
twine upload --repository pypitest dist/pyleri-x.x.x.tar.gz
twine upload --repository pypi dist/pyleri-x.x.x.tar.gz
"""
from setuptools import setup
from pyleri import __version__ as version

try:
    with open('README.md', 'r') as f:
        long_description = f.read()
except IOError:
    long_description = ''

setup(
    name='pyleri',
    packages=['pyleri'],
    version=version,
    description='Python Left-Right Parser',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jeroen van der Heijden',
    author_email='jeroen@cesbit.com',
    url='https://github.com/cesbit/pyleri',
    download_url=(
        'https://github.com/cesbit/'
        'pyleri/tarball/{}'.format(version)),
    keywords=['parser', 'grammar', 'autocompletion'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Linguistic'
    ],
)



"""Repeat class.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
"""
import typing as t
from .elements import Element, NamedElement, c_export, go_export, java_export


class Repeat(NamedElement):

    __slots__ = ('_element', '_min', '_max')

    def __init__(
            self,
            element: Element,
            mi: int = 0,
            ma: t.Optional[int] = None):
        self._element = self._validate_element(element)

        if not isinstance(mi, int) or mi < 0:
            raise TypeError('Repeat(): "mi" must be an integer value larger '
                            'than or equal to 0, got: {mi}'.format(mi))

        self._min = mi

        if not isinstance(ma, (int, type(None))) or \
                (ma is not None and ma < mi):
            raise TypeError('Repeat(): "ma" must be an integer or None value. '
                            'If an integer is specified it must be larger '
                            'than or equal to "mi"')

        self._max = ma

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def _elements(self):
        yield self._element

    def _get_node_result(self, root, tree, rule, s, node):
        pos, i = node.start, 0
        while not self._max or i < self._max:
            is_valid, p = root._walk(
                self._element,
                pos,
                node.children,
                rule,
                i < self._min)
            if not is_valid:
                break
            pos = p
            i += 1

        is_valid = i >= self._min
        if is_valid:
            root._append_tree(tree, node, pos)

        return is_valid, pos

    def _run_export_js(self, js_indent, indent, classes, cname):
        return 'Repeat({}, {}, {})'.format(
            self._element._export_js(js_indent, indent, classes, cname),
            self._min,
            self._max or 'undefined')

    def _run_export_py(self, py_indent, indent, classes):
        return 'Repeat({}, {}, {})'.format(
            self._element._export_py(py_indent, indent, classes),
            self._min,
            self._max or 'None')

    @c_export
    def _run_export_c(self, c_indent, indent, enums, gid):
        # If the repeat is used as a duplication we can use the duplication
        # which is supported by libcleri
        if hasattr(self._element, 'name') \
                and self._min == 1 \
                and self._max == 1:
            return 'cleri_dup({}, {})'.format(
                gid,
                self._element.name)
        return 'cleri_repeat({}, {}, {}, {})'.format(
            gid,
            self._element._export_c(c_indent, indent, enums),
            self._min,
            self._max or 0)

    @go_export
    def _run_export_go(self, go_indent, indent, enums, gid):
        return 'goleri.NewRepeat({}, {}, {}, {})'.format(
            gid,
            self._element._export_go(go_indent, indent, enums),
            self._min,
            self._max or '0')

    @java_export
    def _run_export_java(self, java_indent, indent, enums, classes, gid):
        return 'new Repeat({}{}, {}, {})'.format(
            '' if gid is None else 'Ids.{}, '.format(gid),
            self._element._export_java(java_indent, indent, enums, classes),
            self._min,
            self._max or 'null')
